"""Instruction Selector module."""

from typing import Dict, Any, List, Optional, Union
from ..agent_config import AgentConfig
from .isa_manager import Instruction

class InstructionSelector:
    """Selects appropriate instructions for agent tasks."""

    def __init__(self, config: AgentConfig):
        """Initialize instruction selector."""
        self.config = config
        self.optimization_strategy = config.domain_config.get('optimization_strategy', 'default')
        self.reward_history = []

    async def initialize(self):
        """Initialize the instruction selector."""
        # Nothing to initialize for now
        return self

    def select_instructions(self, input_data: Dict[str, Any], available_instructions: Dict[str, Instruction]) -> List[str]:
        """Select appropriate instructions based on input data.
        
        Args:
            input_data: Input data for instruction selection
            available_instructions: Dictionary of available instructions
            
        Returns:
            List of instruction IDs to execute
        """
        if not available_instructions:
            return []
            
        # For now, return all instruction IDs in order
        return list(available_instructions.keys())
        
    def get_reward(self, result: Dict[str, Any]) -> float:
        """Get reward for instruction execution result.
        
        Args:
            result: Result from instruction execution
            
        Returns:
            Reward value between 0 and 1
        """
        # For now, return a fixed reward if execution was successful
        reward = 1.0 if result.get('status') == 'success' else 0.0
        self.reward_history.append(reward)
        return reward 