"""Evaluation module for ELL2A."""

from datetime import datetime
from typing import Any, Dict, List, Optional, Union, Callable, Type
from pydantic import BaseModel, ConfigDict, Field
import numpy as np

class EvaluationRun(BaseModel):
    """Single evaluation run."""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    timestamp: datetime = Field(default_factory=datetime.now)
    input_data: Dict[str, Any] = Field(default_factory=dict)
    output_data: Dict[str, Any] = Field(default_factory=dict)
    parameters: Dict[str, Any] = Field(default_factory=dict)
    metrics: Dict[str, Any] = Field(default_factory=dict)
    status: str = "pending"
    error: Optional[str] = None
    n_evals: int = 10
    samples_per_datapoint: int = 2

class Evaluation(BaseModel):
    """Evaluation class for ELL2A."""
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    name: str
    description: Optional[str] = None
    parameters: Dict[str, Any] = Field(default_factory=dict)
    runs: List[EvaluationRun] = Field(default_factory=list)
    criterion: Optional[Callable[[Dict[str, Any]], bool]] = None
    n_evals: int = 10
    samples_per_datapoint: int = 2
    metrics: Dict[str, Callable[[Dict[str, Any], Any], float]] = Field(default_factory=dict)
    
    async def _process_single(
        self,
        data_point: Dict[str, Any],
        lmp: Type,
        lmp_params: Dict[str, Any],
        required_params: bool = True
    ) -> List[Any]:
        """Process a single data point.
        
        Args:
            data_point: Input data point
            lmp: Language model provider
            lmp_params: Parameters for the language model provider
            required_params: Whether parameters are required
            
        Returns:
            List of results
            
        Raises:
            ValueError: If input data is invalid or API error occurs
            KeyError: If required input keys are missing
        """
        # Validate input data
        if not isinstance(data_point, dict):
            raise ValueError(f"Input data must be a dictionary, got {type(data_point)}")
        
        if required_params:
            if not data_point:
                raise ValueError("Input data is required")
            if "input" not in data_point:
                raise KeyError("Missing 'input' key in data point")
            if not isinstance(data_point["input"], dict):
                raise ValueError(f"Input value must be a dictionary, got {type(data_point['input'])}")
        
        results = []
        for _ in range(self.samples_per_datapoint):
            try:
                # Create LMP instance with parameters
                if "input" in data_point and isinstance(data_point["input"], dict) and "param" in data_point["input"]:
                    result = await lmp(
                        param=data_point["input"]["param"],
                        api_params=lmp_params
                    )
                else:
                    result = await lmp(api_params=lmp_params)
                results.append(result)
            except ValueError as e:
                # Re-raise ValueError with original message
                raise ValueError(str(e))
            except Exception as e:
                if required_params:
                    raise ValueError(str(e))
                results.append(str(e))
        return results
    
    async def run(
        self,
        lmp: Type,
        data: Optional[List[Dict[str, Any]]] = None,
        n_workers: int = 1,
        verbose: bool = False,
        **kwargs
    ) -> EvaluationRun:
        """Run evaluation.
        
        Args:
            lmp: Language model provider
            data: Input data for evaluation
            n_workers: Number of workers for parallel processing
            verbose: Whether to show verbose output
            **kwargs: Additional arguments
            
        Returns:
            Evaluation run
        """
        # Create evaluation run
        run = EvaluationRun(
            input_data={"lmp": lmp.__name__, "data": data},
            parameters=kwargs,
            timestamp=datetime.now(),
            n_evals=self.n_evals,
            samples_per_datapoint=self.samples_per_datapoint,
            metrics={}  # Initialize metrics dictionary
        )
        
        try:
            # Process input data
            results = []
            data_points = []  # Keep track of data points for each result
            for _ in range(self.n_evals):
                if data:
                    for data_point in data:
                        result = await self._process_single(data_point, lmp, kwargs.get("lmp_params", {}))
                        results.extend(result)
                        data_points.extend([data_point] * len(result))  # Add data point for each result
                else:
                    # Single run without data
                    result = await self._process_single({}, lmp, kwargs.get("lmp_params", {}))
                    results.extend(result)
                    data_points.extend([{}] * len(result))  # Add empty data point for each result
            
            run.output_data["results"] = results
            
            # Initialize metric arrays
            for metric_name in self.metrics:
                run.metrics[metric_name] = []
            
            # Calculate metrics for each result
            for data_point, result in zip(data_points, results):
                for metric_name, metric_fn in self.metrics.items():
                    try:
                        value = metric_fn(data_point, result)  # Pass the correct data point
                        run.metrics[metric_name].append(value)
                    except Exception as e:
                        if verbose:
                            print(f"Error calculating metric {metric_name} for result: {e}")
            
            # Convert metric lists to numpy arrays
            for metric_name in self.metrics:
                if run.metrics[metric_name]:
                    run.metrics[metric_name] = np.array(run.metrics[metric_name])
                else:
                    run.metrics[metric_name] = np.array([0.0])  # Default value if no metrics were calculated
            
            # Apply criterion if provided
            if self.criterion:
                run.metrics["criterion_met"] = self.criterion(run.output_data)
            
            run.status = "completed"
            
        except Exception as e:
            run.status = "failed"
            run.error = str(e)
            if verbose:
                print(f"Error during evaluation: {e}")
        
        # Add run to list
        self.runs.append(run)
        
        return run 