"""Simple decorator for Language Model Providers."""

from functools import wraps
from typing import Any, Callable, Dict, Optional, Union
from ..types.message import Message, MessageRole, MessageType, ContentBlock
from ..integration import ELL2AIntegration

def simple(
    model: str = "gpt-4",
    temperature: float = 0.7,
    max_tokens: int = 2000,
    n: int = 1,
    client: Optional[Any] = None,
    store: Union[bool, str] = False,
    **kwargs
) -> Callable:
    """Simple decorator for Language Model Providers.
    
    Args:
        model: Model name
        temperature: Temperature for sampling
        max_tokens: Maximum number of tokens
        n: Number of completions
        client: Custom client instance
        store: Whether to store results and where
        **kwargs: Additional arguments
        
    Returns:
        Decorated function
    """
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(*args, **func_kwargs):
            # Get ELL2A instance
            ell2a = ELL2AIntegration()
            
            # Create message from function result
            result = await func(*args, **func_kwargs)
            metadata = {
                "model": model,
                "temperature": temperature,
                "max_tokens": max_tokens,
                "n": n,
                "function": func.__name__,
                "store": store,
                **kwargs
            }
            
            # Add client to metadata if provided
            if client:
                metadata["client"] = client
            
            message = Message(
                role=MessageRole.USER,
                content=str(result),
                type=MessageType.TEXT,
                metadata=metadata
            )
            
            # Process message and return its content
            response = await ell2a.process_message(message)
            
            # Return the Message object directly
            return response
        
        return wrapper
    
    return decorator 